/**
 * 
 */
package simulations;

import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;

import com.opencsv.CSVReader;
import com.opencsv.CSVWriter;

import ec.util.MersenneTwisterFast;
import mechanisms.TradingMechanism;
import sim.engine.Schedule;
import sim.engine.SimState;
import traders.SimpleTrader;

/**
 * @author
 *
 */
public abstract class WineSimulation extends SimState implements Serializable, Runnable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 7465578243254838229L;
	public static final double START_VALUE = 100;
	public static final double MIN_VALUE = 25;
	public static final int NUMBER_SOBOL = 10000;
	public static final int NUMBER_SOBOL_LOOPS = 16;
	public static final int LOOPS = 240;
	
	public String simulationName;
	
	protected Thread thread;
	protected int threadID;
	
	//single parameters
	public int simID;
	public int parameterID;
	public int loopID;
	
	public CSVReader reader;
	
	
	//to register parameters
	public ArrayList<double[]> parameters;
	//to register transactions
	public ArrayList<double[]> transactions;
	
	// traders
	public ArrayList<SimpleTrader> buyers;
	public ArrayList<SimpleTrader> sellers;
	
	public static final int AUCTION_MECHANISM_VARIANT = 2;
		
	
	public static final String CSV_FOLDER = "csv/";
	//BASE_x depending on AUCTION_MECHANISM_VARIANT
	public static final String RESULTS_FOLDER = "D:/SIM_PARALLEL/BASE_2/";
	
	public int currentId;
	
	public int numberBuyers;
	public int numberSellers;
	
	public double muBuyers;
	public double sigmaBuyers;
	public double muSellers;
	public double sigmaSellers;
	
	public double lambdaInactivation;
	public double lambdaPlaceOrder;
	public double lambdaRemoveOrder;
	
	public double provision;
	
	//ranges
	public int[] numberBuyersRange = new int[]{1000,5000};
	public int[] numberSellersRange = new int[]{1000,5000};
	
	public double[] muBuyersRange = new double[]{0.0,0.5};
	public double[] sigmaBuyersRange = new double[]{0.0,0.5};
	public double[] muSellersRange = new double[]{0.0,0.5};
	public double[] sigmaSellersRange = new double[]{0.0,0.5};
	
	public double[] lambdaInactivationRange = new double[]{0.0,0.01};
	public double[] lambdaPlaceOrderRange = new double[]{0.01,0.2};
	public double[] lambdaRemoveOrderRange = new double[]{0.01,0.2};
	
	public double[] provisionRange = new double[]{0.0,0.3};
	
	//arrays
	public int[] numberBuyersArray = new int[NUMBER_SOBOL];
	public int[] numberSellersArray = new int[NUMBER_SOBOL];
	
	public double[] muBuyersArray = new double[NUMBER_SOBOL];
	public double[] sigmaBuyersArray = new double[NUMBER_SOBOL];
	public double[] muSellersArray = new double[NUMBER_SOBOL];
	public double[] sigmaSellersArray = new double[NUMBER_SOBOL];
	
	public double[] lambdaInactivationArray = new double[NUMBER_SOBOL];
	public double[] lambdaPlaceOrderArray = new double[NUMBER_SOBOL];
	public double[] lambdaRemoveOrderArray = new double[NUMBER_SOBOL];
	
	public double[] provisionArray = new double[NUMBER_SOBOL];
	
	
	//random generators
	public long seed;
	public MersenneTwisterFast seedGenerator;
	public MersenneTwisterFast randomBuyers;
	public MersenneTwisterFast randomSellers;
	public MersenneTwisterFast randomInactivate;
	public MersenneTwisterFast randomPlace;
	public MersenneTwisterFast randomWithdraw;
	
	public MersenneTwisterFast randomShopSelection;
	
	//trading mechanisms
	public TradingMechanism tradingMechanism;
	
	public void start(){
		super.start();
		scheduleAll();
	}
	
	public abstract void scheduleAll();
	
	//setup random generators
	public void setUpRandom(){
		seedGenerator = new MersenneTwisterFast(seed);
		randomBuyers = new MersenneTwisterFast(seedGenerator.nextLong());
		randomSellers  = new MersenneTwisterFast(seedGenerator.nextLong());
		randomInactivate = new MersenneTwisterFast(seedGenerator.nextLong());
		randomPlace = new MersenneTwisterFast(seedGenerator.nextLong());
		randomWithdraw = new MersenneTwisterFast(seedGenerator.nextLong());
		randomShopSelection = new MersenneTwisterFast(seed);
	}
	
	
	public void addTransaction(int buyerId, int sellerId, double price){
		transactions.add(new double[]{this.parameterID, loopID, this.schedule.getSteps(), buyerId, sellerId, price});
	}
	/**
	 * @param seed
	 */
	public WineSimulation(long seed) {
		super(seed);
		this.seed = seed;
		this.buyers = new ArrayList<SimpleTrader>();
		this.sellers = new ArrayList<SimpleTrader>();
		// TODO Auto-generated constructor stub
	}

	/**
	 * @param random
	 */
	public WineSimulation(MersenneTwisterFast random) {
		super(random);
		// TODO Auto-generated constructor stub
	}

	/**
	 * @param random
	 * @param schedule
	 */
	public WineSimulation(MersenneTwisterFast random, Schedule schedule) {
		super(random, schedule);
		// TODO Auto-generated constructor stub
	}

	/**
	 * @param seed
	 * @param schedule
	 */
	public WineSimulation(long seed, Schedule schedule) {
		super(seed, schedule);
		// TODO Auto-generated constructor stub
	}

	
	public void finish(){
		super.finish();
	}
	
	public void writeTransactions(String fileName){
		CSVWriter writer;
		String[] entries = new String[transactions.get(0).length];
		try {
			writer = new CSVWriter(new FileWriter(RESULTS_FOLDER + "t_" + threadID + "_" + fileName + "_" + parameterID + "_transactions.csv"), ';');
			int length;
			length = transactions.size();
			for (int i = 0; i < length; i++) {
				for (int j = 0; j < transactions.get(0).length; j++) {
					entries[j] = String.valueOf(transactions.get(i)[j]);
				}
				 writer.writeNext(entries);
			}
			writer.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		};
	}
	
	public void writeParameters(String fileName){
		CSVWriter writer;
		String[] entries = new String[parameters.get(0).length];
		try {
			writer = new CSVWriter(new FileWriter(RESULTS_FOLDER + "t_" + threadID + "_" + fileName + "_" + parameterID + "_parameters.csv"), ';');
			int length;
			length = parameters.size();
			for (int i = 0; i < length; i++) {
				for (int j = 0; j < parameters.get(0).length; j++) {
					entries[j] = String.valueOf(parameters.get(i)[j]);
				}
				 writer.writeNext(entries);
			}
			writer.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		};
	}
	
	public static void runSimulation (final SimState model, double maxSimTime, boolean echo) {
		model.start();
		long lastStepTime = System.currentTimeMillis();
		long startTime = System.currentTimeMillis();
		long reportEvery = 1;
		boolean moreSteps;
		long step = 0;
		double simTime = 0;
		while (moreSteps = model.schedule.step(model)) {
			simTime = model.schedule.getTime();
			step = model.schedule.getSteps();			
			if (simTime >= maxSimTime)
				break;
			if (step % reportEvery == 0) {
				if (echo)
					System.err.println("Step="+step+" schedule.getTime()="+((long)(simTime*10))/10.0+" speed steps/s="+(reportEvery*1000.0/(System.currentTimeMillis()-lastStepTime)));
				lastStepTime = System.currentTimeMillis();
			}
		}		
		
		if (echo) {
			System.out.flush();
			if (!moreSteps) 
				System.err.print("The simulation has been exhausted!");
			else 
				System.err.print("Simulation completed!");
			System.err.println(" step="+step+" schedule.getTime()="+((long)(simTime*10))/10.0+" speed steps/s="+(reportEvery*1000.0/(System.currentTimeMillis()-startTime)));
		}
		
	}	
	
	public static final int[] simValuesIxs(int[] possibleValues, int simIx) {
        int[] res = new int[possibleValues.length];
        int val = simIx;
        for (int i = 0; i < possibleValues.length; i++) {
            res[i] = val % possibleValues[i];
            val = val / possibleValues[i];
        }
        return res;
    }
	
	public abstract void setUpAll(); 
	
	public abstract void updateParameters(int i);
	
	public void readSobol(){
		int i =0;
		String csvFile = "sobol.csv";
		try {
            reader = new CSVReader(new FileReader(CSV_FOLDER + csvFile), ';');
            reader.readNext(); //because of header
            String[] line;
            while ((line = reader.readNext()) != null) {
                readSobolSingleLine(i,line);
                i++;
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
	}
	
	
	
	public abstract void readSobolSingleLine(int i, String[] line);
	
	public double getSobolNumber(double weigth, double[] range){
		double result = (range[1]-range[0]) * weigth + range[0];
		return result;
	}
	
	
	public int getSobolNumber(double weigth, int[] range){
		double result = (range[1]-range[0]) * weigth + range[0];
		return (int) Math.rint(result);
	}
		
	@Override
	public void run(){
		System.out.println("Running " +  threadID );
		parameters = new ArrayList<double[]>();
		System.out.println("START at: " + new java.util.Date() + " Thread ID = " + threadID);
		int portion = 500;
		int startSim = threadID * portion;
		int endSim = startSim + portion;
		readSobol();
		//int portion = 100;
		//int startSim = threadID * (400+portion) + 400;
		//int endSim = startSim + portion;
		simID = 0;
		for (int i=startSim;i<endSim;i++) {
			setUpRandom();
			currentId = 0;
			parameterID = i;
			updateParameters(i);
			transactions = new ArrayList<double[]>();
			for (int j = 0; j < NUMBER_SOBOL_LOOPS; j++) {
				 loopID = j;
				 setUpAll();
				 WineSimulation.runSimulation(this, LOOPS-1, false);
				 finish();
				 kill();
				//temporary
				 if(simID %100 ==0){
					 System.out.println("ROUND: " + simID + " at: " + new java.util.Date() + " Thread ID = " + threadID);
				 }
				 simID++;
			 } //for
			writeTransactions(simulationName);
		} //for loop
		writeParameters(simulationName);
		System.out.println("STOP at: " + new java.util.Date() + " Thread ID = " + threadID);
	}
	
	public void startThread(){
		System.out.println("Starting " +  threadID );
	      if (thread == null) {
	         thread = new Thread (this);
	         thread.setPriority(Thread.MAX_PRIORITY);
	         thread.start ();
	      }
	}
	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// TODO Auto-generated method stub

	}

}
